import cv2
import pandas as pd
import numpy as np
from pathlib import Path
from typing import Tuple, Union, List
from torch.utils.data import Dataset, DataLoader
import torch
from torch import nn
from concurrent.futures import ThreadPoolExecutor, as_completed
from collections import namedtuple
from transformers import OneFormerProcessor, OneFormerForUniversalSegmentation
from transformers import AutoImageProcessor, Mask2FormerForUniversalSegmentation
import os
from tqdm.auto import tqdm
import json
from collections import defaultdict
from tqdm.contrib.concurrent import thread_map


# a label and all meta information
Label = namedtuple( 'Label' , [

            'name'        , # The identifier of this label, e.g. 'car', 'person', ... .
                            # We use them to uniquely name a class

            'id'          , # An integer ID that is associated with this label.
                            # The IDs are used to represent the label in ground truth images
                            # An ID of -1 means that this label does not have an ID and thus
                            # is ignored when creating ground truth images (e.g. license plate).
                            # Do not modify these IDs, since exactly these IDs are expected by the
                            # evaluation server.

            'trainId'     , # Feel free to modify these IDs as suitable for your method. Then create
                            # ground truth images with train IDs, using the tools provided in the
                            # 'preparation' folder. However, make sure to validate or submit results
                            # to our evaluation server using the regular IDs above!
                            # For trainIds, multiple labels might have the same ID. Then, these labels
                            # are mapped to the same class in the ground truth images. For the inverse
                            # mapping, we use the label that is defined first in the list below.
                            # For example, mapping all void-type classes to the same ID in training,
                            # might make sense for some approaches.
                            # Max value is 255!

            'category'    , # The name of the category that this label belongs to

            'categoryId'  , # The ID of this category. Used to create ground truth images
                            # on category level.

            'hasInstances', # Whether this label distinguishes between single instances or not

            'ignoreInEval', # Whether pixels having this class as ground truth label are ignored
                            # during evaluations or not

            'color'       , # The color of this label
            ] )
        
def create_cityscapes_label_colormap():
        """Creates a label colormap used in CITYSCAPES segmentation benchmark.
        Returns:
            A colormap for visualizing segmentation results.
        """
        
        labels = [
        #       name                     id    trainId   category            catId     hasInstances   ignoreInEval   color
        Label(  'unlabeled'            ,  0 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'ego vehicle'          ,  1 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'rectification border' ,  2 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'out of roi'           ,  3 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'static'               ,  4 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'dynamic'              ,  5 ,      255 , 'void'            , 0       , False        , True         , (111, 74,  0) ),
        Label(  'ground'               ,  6 ,      255 , 'void'            , 0       , False        , True         , ( 81,  0, 81) ),
        Label(  'road'                 ,  7 ,        0 , 'flat'            , 1       , False        , False        , (128, 64,128) ),
        Label(  'sidewalk'             ,  8 ,        1 , 'flat'            , 1       , False        , False        , (244, 35,232) ),
        Label(  'parking'              ,  9 ,      255 , 'flat'            , 1       , False        , True         , (250,170,160) ),
        Label(  'rail track'           , 10 ,      255 , 'flat'            , 1       , False        , True         , (230,150,140) ),
        Label(  'building'             , 11 ,        2 , 'construction'    , 2       , False        , False        , ( 70, 70, 70) ),
        Label(  'wall'                 , 12 ,        3 , 'construction'    , 2       , False        , False        , (102,102,156) ),
        Label(  'fence'                , 13 ,        4 , 'construction'    , 2       , False        , False        , (190,153,153) ),
        Label(  'guard rail'           , 14 ,      255 , 'construction'    , 2       , False        , True         , (180,165,180) ),
        Label(  'bridge'               , 15 ,      255 , 'construction'    , 2       , False        , True         , (150,100,100) ),
        Label(  'tunnel'               , 16 ,      255 , 'construction'    , 2       , False        , True         , (150,120, 90) ),
        Label(  'pole'                 , 17 ,        5 , 'object'          , 3       , False        , False        , (153,153,153) ),
        Label(  'polegroup'            , 18 ,      255 , 'object'          , 3       , False        , True         , (153,153,153) ),
        Label(  'traffic light'        , 19 ,        6 , 'object'          , 3       , False        , False        , (250,170, 30) ),
        Label(  'traffic sign'         , 20 ,        7 , 'object'          , 3       , False        , False        , (220,220,  0) ),
        Label(  'vegetation'           , 21 ,        8 , 'nature'          , 4       , False        , False        , (107,142, 35) ),
        Label(  'terrain'              , 22 ,        9 , 'nature'          , 4       , False        , False        , (152,251,152) ),
        Label(  'sky'                  , 23 ,       10 , 'sky'             , 5       , False        , False        , ( 70,130,180) ),
        Label(  'person'               , 24 ,       11 , 'human'           , 6       , True         , False        , (220, 20, 60) ),
        Label(  'rider'                , 25 ,       12 , 'human'           , 6       , True         , False        , (255,  0,  0) ),
        Label(  'car'                  , 26 ,       13 , 'vehicle'         , 7       , True         , False        , (  0,  0,142) ),
        Label(  'truck'                , 27 ,       14 , 'vehicle'         , 7       , True         , False        , (  0,  0, 70) ),
        Label(  'bus'                  , 28 ,       15 , 'vehicle'         , 7       , True         , False        , (  0, 60,100) ),
        Label(  'caravan'              , 29 ,      255 , 'vehicle'         , 7       , True         , True         , (  0,  0, 90) ),
        Label(  'trailer'              , 30 ,      255 , 'vehicle'         , 7       , True         , True         , (  0,  0,110) ),
        Label(  'train'                , 31 ,       16 , 'vehicle'         , 7       , True         , False        , (  0, 80,100) ),
        Label(  'motorcycle'           , 32 ,       17 , 'vehicle'         , 7       , True         , False        , (  0,  0,230) ),
        Label(  'bicycle'              , 33 ,       18 , 'vehicle'         , 7       , True         , False        , (119, 11, 32) ),
        Label(  'license plate'        , -1 ,       -1 , 'vehicle'         , 7       , False        , True         , (  0,  0,142) ),
        ]
        return labels

def create_mapillary_vistas_label_colormap():
    """Creates a label colormap used in Mapillary Vistas segmentation benchmark.
    Returns:
        A list of labels for visualizing segmentation results.
    """

    labels = [
        #       name                id  trainId  category            catId  hasInstances  ignoreInEval  color
        Label('Bird',               0,      0, 'animal',             0,      True,         False,        (165, 42, 42)),
        Label('Ground Animal',      1,      1, 'animal',             0,      True,         False,        (0, 192, 0)),
        Label('Curb',               2,      2, 'construction',       1,      False,        False,        (196, 196, 196)),
        Label('Fence',              3,      3, 'construction',       1,      False,        False,        (190, 153, 153)),
        Label('Guard Rail',         4,      4, 'construction',       1,      False,        False,        (180, 165, 180)),
        Label('Barrier',            5,      5, 'construction',       1,      False,        False,        (102, 102, 156)),
        Label('Wall',               6,      6, 'construction',       1,      False,        False,        (102, 102, 156)),
        Label('Bike Lane',          7,      7, 'flat',               2,      False,        False,        (128, 64, 255)),
        Label('Crosswalk - Plain',  8,      8, 'flat',               2,      False,        False,        (140, 140, 200)),
        Label('Curb Cut',           9,      9, 'flat',               2,      False,        False,        (170, 170, 170)),
        Label('Parking',           10,     10, 'flat',               2,      False,        False,        (250, 170, 160)),
        Label('Pedestrian Area',   11,     11, 'flat',               2,      False,        False,        (96, 96, 96)),
        Label('Rail Track',        12,     12, 'flat',               2,      False,        False,        (230, 150, 140)),
        Label('Road',              13,     13, 'flat',               2,      False,        False,        (128, 64, 128)),
        Label('Service Lane',      14,     14, 'flat',               2,      False,        False,        (110, 110, 110)),
        Label('Sidewalk',          15,     15, 'flat',               2,      False,        False,        (244, 35, 232)),
        Label('Bridge',            16,     16, 'construction',       1,      False,        False,        (150, 100, 100)),
        Label('Building',           17,     17, 'construction',       1,      False,        False,        (70, 70, 70)),
        Label('Tunnel',             18,     18, 'construction',       1,      False,        False,        (150, 120, 90)),
        Label('Person',             19,     19, 'human',              3,      True,         False,        (220, 20, 60)),
        Label('Bicyclist',          20,     20, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Motorcyclist',       21,     21, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Other Rider',        22,     22, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Lane Marking - Crosswalk', 23, 23, 'marking',         4,      False,        False,        (200, 128, 128)),
        Label('Lane Marking - General', 24, 24, 'marking',           4,      False,        False,        (255, 255, 255)),
        Label('Mountain',           25,     25, 'nature',             5,      False,        False,        (64, 170, 64)),
        Label('Sand',               26,     26, 'nature',             5,      False,        False,        (230, 160, 50)),
        Label('Sky',                27,     27, 'nature',             5,      False,        False,        (70, 130, 180)),
        Label('Snow',               28,     28, 'nature',             5,      False,        False,        (190, 255, 255)),
        Label('Terrain',            29,     29, 'nature',             5,      False,        False,        (152, 251, 152)),
        Label('Vegetation',         30,     30, 'nature',             5,      False,        False,        (107, 142, 35)),
        Label('Water',              31,     31, 'nature',             5,      False,        False,        (0, 170, 30)),
        Label('Banner',             32,     32, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Bench',              33,     33, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Bike Rack',          34,     34, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Billboard',          35,     35, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Catch Basin',        36,     36, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('CCTV Camera',        37,     37, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Fire Hydrant',       38,     38, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Junction Box',       39,     39, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Mailbox',            40,     40, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Manhole',            41,     41, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Phone Booth',        42,     42, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Pothole',            43,     43, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Street Light',       44,     44, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Traffic Cone',       45,     45, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Traffic Device',     46,     46, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Traffic Light',      47,     47, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Traffic Sign',       48,     48, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Traffic Sign Frame', 49,     49, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Trash Can',          50,     50, 'object',             6,      False,        False,        (255, 255, 128)),
        Label('Bicycle',            51,     51, 'vehicle',            7,      True,         False,        (119, 11, 32)),
        Label('Boat',               52,     52, 'vehicle',            7,      True,         False,        (0, 0, 142)),
        Label('Bus',                53,     53, 'vehicle',            7,      True,         False,        (0, 60, 100)),
        Label('Car',                54,     54, 'vehicle',            7,      True,         False,        (0, 0, 142)),
        Label('Caravan',            55,     55, 'vehicle',            7,      True,         False,        (0, 0, 90)),
        Label('Motorcycle',         56,     56, 'vehicle',            7,      True,         False,        (0, 0, 230)),
        Label('On Rails',           57,     57, 'vehicle',            7,      True,         False,        (0, 80, 100)),
        Label('Other Vehicle',      58,     58, 'vehicle',            7,      True,         False,        (128, 64, 128)),
        Label('Trailer',            59,     59, 'vehicle',            7,      True,         False,        (0, 0, 110)),
        Label('Truck',              60,     60, 'vehicle',            7,      True,         False,        (0, 0, 70)),
        Label('Wheeled Slow',       61,     61, 'vehicle',            7,      True,         False,        (0, 0, 192)),
        Label('Car Mount',          62,     62, 'vehicle',            7,      False,        False,        (32, 32, 32)),
        Label('Ego Vehicle',        63,     63, 'vehicle',            7,      False,        False,        (120, 10, 10))
    ]

    return labels

def get_resized_dimensions(image: cv2.Mat, max_size: int = 2048) -> Tuple[int, int]:
    """
    Get the resized dimensions of an image based on a maximum size.

    Args:
        image (cv2.Mat): Input image in OpenCV format.
        max_size (int, optional): Maximum size for the larger side of the image. Defaults to 2048.

    Returns:
        Tuple[int, int]: Tuple containing new height and width.
    """
    height, width = image.shape[:2]

    # Determine the larger side
    larger_side = max(height, width)

    # Check if the larger side exceeds the maximum size
    if larger_side > max_size:
        # Calculate the scaling factor
        scaling_factor = max_size / larger_side

        # Calculate the new dimensions
        new_width = int(width * scaling_factor)
        new_height = int(height * scaling_factor)

        return (new_height, new_width)

    return (height, width)

class ImageDataset(Dataset):
    """
    Custom Dataset class for images.
    """

    def __init__(self, dir_input: Union[str, Path], rgb: bool = True) -> None:
        """
        Initialize the ImageDataset.

        Args:
            dir_input (Union[str, Path]): Input directory containing image files.
            rgb (bool, optional): Flag to convert images to RGB. Defaults to True.
        """
        self.dir_input = Path(dir_input)
        self.image_files = list(self.dir_input.glob("*.jpg"))
        self.rgb = rgb

    def __len__(self) -> int:
        """
        Get the length of the dataset.

        Returns:
            int: Number of image files in the dataset.
        """
        return len(self.image_files)

    def __getitem__(self, idx: int) -> Tuple[str, cv2.Mat, Tuple[int, int]]:
        """
        Get an item from the dataset.

        Args:
            idx (int): Index of the image.

        Returns:
            Tuple[str, cv2.Mat, Tuple[int, int]]: Tuple containing image file path, image data, and original image dimensions.
        """
        image_file = self.image_files[idx]
        img = cv2.imread(str(image_file))
        original_img_shape = get_resized_dimensions(img)
        if self.rgb:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return str(image_file), img, original_img_shape

    def collate_fn(self, data: List[Tuple[str, cv2.Mat, Tuple[int, int]]]) -> Tuple[List[str], List[cv2.Mat], List[Tuple[int, int]]]:
        """
        Custom collate function for the dataset.

        Args:
            data (List[Tuple[str, cv2.Mat, Tuple[int, int]]]): List of tuples containing image file path, image data, and original image dimensions.

        Returns:
            Tuple[List[str], List[cv2.Mat], List[Tuple[int, int]]]: Tuple containing lists of image file paths, image data, and original image dimensions.
        """
        image_files, images, original_img_shape = zip(*data)
        return list(image_files), list(images), list(original_img_shape)

class Segmenter:
    def __init__(self, model_name="facebook/mask2former-swin-tiny-cityscapes-semantic", dataset="cityscapes"):
        """
        Initialize the Segmenter with a model and dataset.

        Args:
            model_name (str): The name of the pre-trained model.
            dataset (str): The name of the dataset.
        """
        self.device = self._get_device()
        self.model, self.processor = self._get_model_processor(model_name)
        self.color_map = self._create_color_map(dataset)
        self.label_map = self._create_label_map(dataset)

    def _get_model_processor(self, model_name):
        """
        Get the model and processor for the given model name.

        Args:
            model_name (str): The name of the pre-trained model.

        Returns:
            Tuple: The model and processor.
        """
        # Add other models in the future
        if "mask2former" in model_name:
            processor = AutoImageProcessor.from_pretrained(model_name)
            model = Mask2FormerForUniversalSegmentation.from_pretrained(model_name).to(self.device)
        return model, processor

    def _create_color_map(self, dataset):
        """
        Create a color map based on the given dataset.

        Args:
            dataset (str): The name of the dataset.

        Returns:
            np.ndarray: A color map.
        """
        if dataset == "cityscapes":
            labels = create_cityscapes_label_colormap()
        elif dataset == "mapillary":
            labels = create_mapillary_vistas_label_colormap()
        else:
            raise ValueError(f"Unknown dataset: {dataset}")

        labels = [label for label in labels if label.trainId != -1]
        train_ids = np.array([label.trainId for label in labels], dtype=np.uint8)
        colors = np.array([label.color for label in labels], dtype=np.uint8)
        max_train_id = np.max(train_ids) + 1
        color_map = np.zeros((max_train_id, 3), dtype=np.uint8)
        color_map[train_ids] = colors

        # Add a train_id_to_name dictionary as an attribute
        self.train_id_to_name = {label.trainId: label.name for label in labels}

        return color_map

    def _create_label_map(self, dataset):
        """
        Create a label map based on the given dataset.

        Args:
            dataset (str): The name of the dataset.

        Returns:
            Dict[Tuple, Label]: A dictionary mapping colors to labels.
        """
        if dataset == "cityscapes":
            labels = create_cityscapes_label_colormap()
        elif dataset == "mapillary":
            labels = create_mapillary_vistas_label_colormap()
        else:
            raise ValueError(f"Unknown dataset: {dataset}")

        color_to_label = {}
        for label in labels:
            color = label.color
            color_to_label[color] = label

        return color_to_label

    def _get_device(self) -> torch.device:
        """
        Get the appropriate device for running the model.

        Returns:
            torch.device: The device to use for running the model.
        """
        if torch.cuda.is_available():
            print("Using GPU")
            return torch.device("cuda")
        else:
            print("Using CPU")
            return torch.device("cpu")

    def _calculate_pixel_ratios(self, segmented_img):
        """
        Calculate pixel ratios for each class in the segmented image.

        Args:
            segmented_img (numpy.ndarray): Segmented image.

        Returns:
            dict: A dictionary with class names as keys and pixel ratios as values.
        """
        unique, counts = np.unique(segmented_img, return_counts=True)
        total_pixels = np.sum(counts)
        pixel_ratios = {self.train_id_to_name[train_id]: count / total_pixels for train_id, count in zip(unique, counts)}

        return pixel_ratios

    def _save_pixel_ratios_as_csv(self, pixel_ratio_dict, dir_output):
        """
        Save pixel ratios dictionary as a CSV file.

        Args:
            pixel_ratio_dict (dict): Dictionary with class names as keys and pixel ratios as values.
            dir_output (pathlib.Path): Output directory path.
        """
        pixel_ratios_df = pd.DataFrame(pixel_ratio_dict).transpose()
        pixel_ratios_df.fillna(0, inplace=True)
        pixel_ratios_df.index.names = ["filename_key"]
        pixel_ratios_df.to_csv(dir_output / "pixel_ratios.csv")

    def _panoptic_segmentation(self, images, original_img_shape):
        """
        Perform panoptic segmentation on the given images.

        Args:
            images (list): List of input images.
            original_img_shape (tuple): Original image shape.

        Returns:
            list: List of panoptic segmentation outputs.
        """
        inputs = self.processor(images=images, task_inputs=["panoptic"], return_tensors="pt").to(self.model.device)
        outputs = self.model(**inputs)
        return self.processor.post_process_panoptic_segmentation(outputs, target_sizes=original_img_shape)

    def _semantic_segmentation(self, images, original_img_shape):
        """
        Perform semantic segmentation on the given images.

        Args:
            images (list): List of input images.
            original_img_shape (tuple): Original image shape.

        Returns:
            tuple: Tuple containing list of semantic segmentation outputs and list of pixel ratios.
        """
        inputs = self.processor(images=images, return_tensors="pt").to(self.model.device)
        with torch.no_grad():
            outputs = self.model(**inputs)
        segmentations = self.processor.post_process_semantic_segmentation(outputs, target_sizes=original_img_shape)

        pixel_ratios = [self._calculate_pixel_ratios(segmentation.cpu().numpy()) for segmentation in segmentations]

        return segmentations, pixel_ratios

    def _trainid_to_color(self, segmented_img):
        """
        Convert segmented image with train IDs to a colored image.

        Args:
            segmented_img (numpy.ndarray): Segmented image with train IDs.

        Returns:
            numpy.ndarray: Colored segmented image.
        """
        colored_img = self.color_map[segmented_img]
        return colored_img

    def _save_panoptic_segmentation_image(self, image_file, img, dir_output, output):
        """
        Save the panoptic segmentation image as a blended image with the original input image.

        Args:
            image_file (str): The input image file path.
            img (np.array): The input image in the format of a NumPy array.
            dir_output (Path): The output directory path to save the blended image.
            output (dict): The output dictionary containing the segmentation data.
        """
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        segmented_img = output["segmentation"].cpu().numpy()
        colored_segmented_img = self._trainid_to_color(segmented_img)
        
        alpha = 0.5
        blend_img = cv2.addWeighted(img, alpha, colored_segmented_img, 1 - alpha, 0)

        output_file = dir_output / Path(image_file).name
        cv2.imwrite(str(output_file), cv2.cvtColor(blend_img, cv2.COLOR_RGB2BGR))
        
    def _save_semantic_segmentation_image(self, image_file, img, dir_output, output):
        """
        Save the semantic segmentation image as a colored segmented image and/or a blended image with the original input image.

        Args:
            image_file (str): The input image file path.
            img (np.array): The input image in the format of a NumPy array.
            dir_output (Path): The output directory path to save the colored segmented and/or blended image.
            output (Tensor): The output tensor containing the semantic segmentation data.
        """
        colored_segmented_img = self._trainid_to_color(output.cpu().numpy())
        alpha = 0.5
        blend_img = cv2.addWeighted(img, alpha, colored_segmented_img, 1 - alpha, 0)

        output_file = dir_output / Path(image_file).name
        
        # Save images based on specified options
        if "segmented_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_colored_segmented.png")), cv2.cvtColor(colored_segmented_img, cv2.COLOR_RGB2BGR))
        if "blend_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_blend.png")), cv2.cvtColor(blend_img, cv2.COLOR_RGB2BGR))

    def _save_segmentation_image(self, task, image_file, img, dir_output, output):
        """
        Save the segmentation image based on the given task.

        Args:
            task (str): The segmentation task, either "panoptic" or "semantic".
            image_file (str): The input image file path.
            img (np.array): The input image in the format of a NumPy array.
            dir_output (Path): The output directory path to save the segmentation images.
            output (dict or Tensor): The output data containing the segmentation results.
        """
        if task == "panoptic":
            self._save_panoptic_segmentation_image(image_file, img, dir_output, output)
        elif task == "semantic":
            self._save_semantic_segmentation_image(image_file, img, dir_output, output)

    def _process_images(self, task, image_files, images, dir_output, pixel_ratio_dict, original_img_shape):
        """
        Process the input images for segmentation and save the output images.

        Args:
            task (str): The segmentation task to perform, either "panoptic" or "semantic".
            image_files (List[str]): The list of file paths of the input images.
            images (List[ndarray]): The list of input images in the form of numpy arrays.
            dir_output (Path): The output directory where the segmented images will be saved.
            pixel_ratio_dict (defaultdict): A dictionary to store the pixel ratios of the segmented images.
            original_img_shape (List[Tuple[int, int]]): The original shapes of the input images.

        Returns:
            None
        """
        outputs = None
        pixel_ratios = None
        if task == "panoptic":
            outputs = self._panoptic_segmentation(images, original_img_shape)
        elif task == "semantic":
            outputs, pixel_ratios = self._semantic_segmentation(images, original_img_shape)

        if outputs is not None:
            for image_file, img, output, pixel_ratio in zip(image_files, images, outputs, pixel_ratios):
                if len(self.save_image_options) > 0:
                    self._save_segmentation_image(task, image_file, img, dir_output, output)
                image_file_key = Path(image_file).stem
                pixel_ratio_dict[image_file_key] = pixel_ratio

    # Modify the segment method inside the Segmenter class
    def segment(self, dir_input: Union[str, Path], dir_image_output: Union[str, Path, None] = None, dir_pixel_ratio_output: Union[str, Path, None] = None, task="semantic", batch_size=1, num_workers=1, save_image_options = ["segmented_image", "blend_image"], pixel_ratio_save_format = ["json", "csv"]):
        """
        Perform segmentation on the input images and save the segmented images.

        Args:
            dir_input (Union[str, Path]): The directory containing the input images.
            dir_image_output (Union[str, Path , None]): The output directory where the segmented images will be saved. Default is None.
            dir_pixel_ratio_output (Union[str, Path, None]): The output directory where the pixel ratio data will be saved. Default is None.
            task (str): The segmentation task to perform, either "panoptic" or "semantic". Default is "semantic".
            batch_size (int): The batch size to use for segmentation. Default is 1.
            num_workers (int): The number of worker threads to use for segmentation. Default is 0.
            save_image_options (List[str]): A list of options for saving the segmented images. Possible options are "segmented_image" and "blend_image". Default is ["segmented_image", "blend_image"].
            pixel_ratio_save_format (List[str]): A list of output formats for the pixel ratio data. Possible options are "json" and "csv". Default is ["json", "csv"].

        Returns:
            None
        """
        # make sure that at least one of dir_image_output and dir_pixel_ratio_output is not None
        if dir_image_output is None and dir_pixel_ratio_output is None:
            raise ValueError("At least one of dir_image_output and dir_pixel_ratio_output must not be None.")
        
        # save_image_options as a property of the class
        self.save_image_options = save_image_options
        
        # make directory
        dir_input = Path(dir_input)
        dir_image_output = Path(dir_image_output)
        dir_image_output.mkdir(parents=True, exist_ok=True)
        dir_pixel_ratio_output = Path(dir_pixel_ratio_output)
        dir_pixel_ratio_output.mkdir(parents=True, exist_ok=True)

        dataset = ImageDataset(dir_input)
        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=dataset.collate_fn, num_workers=num_workers)

        pixel_ratio_dict = defaultdict(dict)

        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = []

            for batch in tqdm(dataloader, desc="Submitting tasks"):
                image_files, images, original_img_shape = batch
                future = executor.submit(self._process_images, task, image_files, images, dir_image_output, pixel_ratio_dict, original_img_shape)
                futures.append(future)

            for completed_future in tqdm(as_completed(futures), total=len(futures), desc="Processing tasks"):
                completed_future.result()

        # Save pixel_ratio_dict as a JSON or CSV file
        if "json" in pixel_ratio_save_format:
            with open(dir_pixel_ratio_output / "pixel_ratios.json", "w") as f:
                json.dump(pixel_ratio_dict, f)
        if "csv" in pixel_ratio_save_format:
            self._save_pixel_ratios_as_csv(pixel_ratio_dict, dir_pixel_ratio_output)
            
    def calculate_pixel_ratio_post_process(self, dir_input, dir_output, pixel_ratio_save_format = ["json", "csv"], num_workers=1):
        """
        Calculates the pixel ratio of different classes present in the segmented images and saves the results in either JSON or CSV format.

        Args:
            dir_input: A string or Path object representing the input directory containing the segmented images.
            dir_output: A string or Path object representing the output directory where the pixel ratio results will be saved.
            pixel_ratio_save_format: A list containing the file formats in which the results will be saved. The allowed file formats are "json" and "csv". The default value is ["json", "csv"].

        Returns:
            None
        """
        def calculate_label_ratios(image, label_map):
            """
            Calculates the pixel ratio of different classes present in a single image.

            Args:
                image: A numpy array representing an image.
                label_map: A dictionary containing the label names and their respective RGB colors.

            Returns:
            A dictionary containing the pixel ratio of different classes in the given image.
            """
            label_ratios = {}
            total_pixels = image.shape[0] * image.shape[1]

            for color, label in label_map.items():
                color_pixels = np.count_nonzero(np.all(image == color, axis=-1))
                label_ratios[label.name] = color_pixels / total_pixels

            return label_ratios

        def process_image_file(image_file, label_map):
            """
            Calculates the pixel ratio of different classes in a single segmented image file.

            Args:
                image_file: A Path object representing the segmented image file.
                label_map: A dictionary containing the label names and their respective RGB colors.

            Returns:
                A tuple containing the image file key and the pixel ratio of different classes in the given image.
            """
            image_file_key = str(Path(image_file).stem).replace("_colored_segmented", "")
            image = cv2.imread(str(image_file))
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            label_ratios = calculate_label_ratios(image, label_map)
            return image_file_key, label_ratios

        def results_to_dataframe(results):
            """
            Converts the results obtained from processing each image file into a Pandas DataFrame.

            Args:
                results: A list of tuples, where each tuple contains the image file key and the pixel ratio of different classes in the corresponding image.

            Returns:
                A Pandas DataFrame containing the pixel ratios of different classes in each image file.
            """
            pixel_ratio_dict = {}

            for image_file_key, label_ratios in results:
                pixel_ratio_dict[str(image_file_key)] = label_ratios

            pixel_ratios_df = pd.DataFrame(pixel_ratio_dict).transpose()
            pixel_ratios_df.fillna(0, inplace=True)
            pixel_ratios_df.index.names = ["filename_key"]

            return pixel_ratios_df
        
        def results_to_nested_dict(results):
            """
            Converts the results obtained from processing each image file into a nested dictionary.

            Args:
                results: A list of tuples, where each tuple contains the image file key and the pixel ratio of different classes in the corresponding image.

            Returns:
                A nested dictionary containing the pixel ratios of different classes in each image file.
            """
            data = {}

            for image_file_key, label_ratios in results:
                image_file_key = str(image_file_key)
                data[image_file_key] = label_ratios

            return data

        # get files
        if isinstance(dir_input, str):
            dir_input = Path(dir_input)

        # Set image file extensions
        image_extensions = ['.jpg', '.png']

        image_files = [file for file in dir_input.rglob('*') if file.suffix.lower() in image_extensions and '_colored_segmented' in file.stem]

        results = thread_map(process_image_file, image_files, [self.label_map] * len(image_files), max_workers=num_workers)

        if "json" in pixel_ratio_save_format:
            json_output_file = Path(dir_output) / 'pixel_ratio.json'
            nested_dict = results_to_nested_dict(results)
            with open(json_output_file, 'w') as f:
                json.dump(nested_dict, f, indent=2) 
                
        if "csv" in pixel_ratio_save_format:
            csv_output_file = Path(dir_output) / 'pixel_ratio.csv'
            df = results_to_dataframe(results)
            df.to_csv(csv_output_file)


import osmnx as ox

def standardize_column_names(df):
    longitude_variants = ['longitude', 'long', 'lon', 'lng', "x"]
    latitude_variants = ['latitude', 'lat', 'lt', "y"]
    # convert all column names to lowercase
    df.columns = [col.lower() for col in df.columns]      
    for col in df.columns:
        if col in longitude_variants:
            df.rename(columns={col: 'longitude'}, inplace=True)
        elif col in latitude_variants:
            df.rename(columns={col: 'latitude'}, inplace=True)
    return df

def create_buffer_gdf(gdf, buffer_distance):
    if gdf.crs == None:
        gdf = gdf.set_crs("EPSG:4326")
    # Project the GeoDataFrame to UTM
    gdf_projected = ox.projection.project_gdf(gdf)
    
    # Buffer the points by buffer_distance
    gdf_projected['geometry'] = gdf_projected.buffer(buffer_distance)
    
    # Project the GeoDataFrame back to EPSG:4326
    gdf = gdf_projected.to_crs('EPSG:4326')
    
    return gdf


import geopandas as gpd
import geopy.distance
import osmnx as ox
import pandas as pd
from shapely.geometry import Point
from tqdm.auto import tqdm
tqdm.pandas()
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
import numpy as np
from pyproj import Transformer
import networkx

class GeoProcessor:
    def __init__(self, gdf, distance=1, grid = False, grid_size = 1, id_columns=[]):
        self.gdf = gdf
        self.distance = distance
        self.processing_functions = {
            'Point': self.process_point,
            'MultiPoint': self.process_multipoint,
            'LineString': self.process_linestring,
            'MultiLineString': self.process_multilinestring,
            'Polygon': self.process_polygon,
            'MultiPolygon': self.process_multipolygon
        }
        self.grid = grid
        self.grid_size = grid_size
        self.utm_crs = None
        self.id_columns = id_columns

    def get_lat_lon(self):
        self.gdf['feature_type'] = self.gdf['geometry'].apply(lambda x: x.geom_type)
        gdf_list = []

        for feature_type, func in self.processing_functions.items():
            sub_gdf = self.gdf[self.gdf['feature_type'] == feature_type]
            if not sub_gdf.empty:
                print(f"Getting longitude and latitude from {feature_type} feature")
                processed_gdf = func(sub_gdf)
                gdf_list.append(processed_gdf)

        result_gdf = pd.concat(gdf_list)
        return result_gdf

    # Define other processing functions (process_point, process_multipoint, etc.) as class methods here
    def process_point(self, gdf):
        gdf['longitude'] = gdf.geometry.x
        gdf['latitude'] = gdf.geometry.y
        return gdf[self.id_columns + ['longitude', 'latitude']]

    def process_multipoint(self, gdf):
        # Explode the multipoint into individual points
        gdf = gdf.explode('geometry').reset_index(drop=True)

        # Call process_point on the exploded GeoDataFrame
        return self.process_point(gdf)

    def process_linestring(self, gdf):
        # Project the LineStrings to UTM
        gdf_utm = ox.projection.project_gdf(gdf)
        self.utm_crs = gdf_utm.crs

        # Use osmnx.utils_geo.interpolate_points function to interpolate points along LineStrings
        gdf_utm['sample_points'] = gdf_utm['geometry'].progress_apply(lambda geom: list(ox.utils_geo.interpolate_points(geom, dist=self.distance)), desc="Interpolating Points")
        gdf_utm = gdf_utm.explode('sample_points').reset_index(drop=True)

        # Convert the UTM points to latitude and longitude
        gdf['longitude'], gdf['latitude'] = zip(*self.utm_to_lat_lon(gdf_utm['sample_points'].apply(lambda p: (p.x, p.y)).tolist(), self.utm_crs))
        return gdf[self.id_columns + ['longitude', 'latitude']]

    def process_multilinestring(self, gdf):
        # Explode the multilinestring into individual linestrings
        gdf = gdf.explode('geometry').reset_index(drop=True)

        # Call process_linestring on the exploded GeoDataFrame
        return self.process_linestring(gdf)

    def get_street_points(self, polygon):
        graph = ox.graph_from_polygon(polygon, network_type='all')
        _, edges = ox.graph_to_gdfs(graph)
        # Project the street network to UTM
        edges_utm = ox.projection.project_gdf(edges)
        utm_crs = edges_utm.crs

        # Use osmnx.utils_geo.interpolate_points function to interpolate points along LineStrings
        edges_utm['sample_points'] = edges_utm['geometry'].apply(lambda geom: list(ox.utils_geo.interpolate_points(geom, dist=self.distance)))
        edges_utm = edges_utm.explode('sample_points').reset_index(drop=True)
        return edges_utm['sample_points'].tolist(), utm_crs

    def create_point_grid(self, polygon, grid_size, crs="EPSG:4326"):
        """
        Create a point grid within the bounding box of the input GeoDataFrame with the given grid size in meters.

        Args:
            polygon (geopandas.GeoDataFrame): The input GeoDataFrame to get the bounding box from.
            grid_size (float): The size of the grid in meters.
            crs (str, optional): The coordinate reference system for the points. Defaults to "EPSG:4326".

        Returns:
            list: A list of shapely Point objects in UTM coordinates.
        """
        # Compute the bounding box coordinates
        west, south, east, north = polygon.bounds

        # Convert the bounding box coordinates to UTM for accurate distance calculations
        polygon_geom = ox.utils_geo.bbox_to_poly(north, south, east, west)
        bounding_box = gpd.GeoDataFrame({"geometry": [polygon_geom]}, crs=crs)
        bounding_box_utm = ox.projection.project_gdf(bounding_box)
        utm_crs = bounding_box_utm.crs
        west_utm, south_utm, east_utm, north_utm = bounding_box_utm.total_bounds

        # Create the point grid using NumPy
        x_coords = np.arange(west_utm, east_utm, grid_size)
        y_coords = np.arange(south_utm, north_utm, grid_size)
        x_grid, y_grid = np.meshgrid(x_coords, y_coords)
        points = np.vstack((x_grid.ravel(), y_grid.ravel())).T
        # Create a GeoSeries of points in UTM coordinates
        points_utm = gpd.GeoSeries((Point(xy) for xy in points), crs=utm_crs)

        # Return the points as coordinates (x, y) in UTM coordinates
        return [(point.x, point.y) for point in points_utm], utm_crs
    
    def utm_to_lat_lon(self, utm_points, utm_crs):
        transformer = Transformer.from_crs(utm_crs, "EPSG:4326", always_xy=True)
        lat_lon_points = [transformer.transform(x, y) for x, y in tqdm(utm_points, desc="Converting UTM to Lat/Lon")]
        return lat_lon_points

    def process_polygon(self, gdf):
        batch_size = 100  # Modify this to a suitable value
        num_batches = (len(gdf) + batch_size - 1) // batch_size
        failed_geoms = []
        results = []

        for i in tqdm(range(num_batches), desc=f"Processing polgyon by batch size {min(batch_size, len(gdf))}"):
            with ProcessPoolExecutor() as executor:
                batch_futures = {}
                for geom in gdf.geometry.iloc[i*batch_size : (i+1)*batch_size]:
                    if self.grid == False:
                        future = executor.submit(self.get_street_points, geom)
                    else:
                        future = executor.submit(self.create_point_grid, geom, self.grid_size)
                    batch_futures[future] = geom
                
                for future in tqdm(as_completed(batch_futures.keys()), total=len(batch_futures), desc=f"Processing polygon for batch #{i+1}"):
                    geom = batch_futures[future]
                    try:
                        result = future.result()
                        results.append(result)
                    except (ValueError, networkx.exception.NetworkXPointlessConcept):
                        tqdm.write("Found no graph nodes within the polygon or connectivity is undefined for the null graph. Store the polygon as a failed geom and retry later")
                        failed_geoms.append(geom)

        if len(failed_geoms) > 0:
            print("Retrying failed geoms")
            with ProcessPoolExecutor() as executor:
                retry_futures = {}
                for geom in tqdm(failed_geoms, desc="Preparing Failed Geoms"):
                    future = executor.submit(self.create_point_grid, geom, self.grid_size)
                    retry_futures[future] = geom
                
                for future in tqdm(as_completed(retry_futures.keys()), total=len(retry_futures), desc="Processing Failed Geoms"):
                    result = future.result()
                    results.append(result)

        gdf['street_points'] = [result[0] for result in results]
        self.utm_crs = results[0][1]

        gdf = gdf.explode('street_points').reset_index(drop=True)

        # Convert the UTM points to latitude and longitude
        gdf['longitude'], gdf['latitude'] = zip(*self.utm_to_lat_lon(gdf['street_points'].tolist(), self.utm_crs))

        return gdf[self.id_columns + ['longitude', 'latitude']]



    def process_multipolygon(self, gdf):
        # Explode the multipolygon into individual polygons
        gdf = gdf.explode('geometry').reset_index(drop=True)

        # Call process_polygon on the exploded GeoDataFrame
        return self.process_polygon(gdf)



import re
from datetime import datetime
import requests
import random

def _panoids_url(lat, lon):
    url = "https://maps.googleapis.com/maps/api/js/GeoPhotoService.SingleImageSearch?pb=!1m5!1sapiv3!5sUS!11m2!1m1!1b0!2m4!1m2!3d{0:}!4d{1:}!2d50!3m10!2m2!1sen!2sGB!9m1!1e2!11m4!1m3!1e2!2b1!3e2!4m10!1e1!1e2!1e3!1e4!1e8!1e6!5m1!1e2!6m1!1e2&callback=_xdc_._v2mub5"
    return url.format(lat, lon)

def _panoids_data(lat, lon, proxies):
    url = _panoids_url(lat, lon)
    while True:
        proxy = random.choice(proxies)
        try:
            resp = requests.get(url, proxies=proxy, timeout=5)
            return resp.text
        except Exception as e:
            print(f"Proxy {proxy} is not working. Exception: {e}")
            continue

def panoids(lat, lon, proxies, closest=False, disp=False):
    resp = _panoids_data(lat, lon, proxies)

    # Get all the panorama ids and coordinates
    pans = re.findall('\[[0-9]+,"(.+?)"\].+?\[\[null,null,(-?[0-9]+.[0-9]+),(-?[0-9]+.[0-9]+)', resp)
    pans = [{"panoid": p[0], "lat": float(p[1]), "lon": float(p[2])} for p in pans]

    # Get all the dates
    dates = re.findall('([0-9]?[0-9]?[0-9])?,?\[(20[0-9][0-9]),([0-9]+)\]', resp)
    dates = [list(d)[1:] for d in dates]

    if len(dates) > 0:
        # Convert all values to integers
        dates = [[int(v) for v in d] for d in dates]

        # Make sure the month value is between 1-12
        dates = [d for d in dates if d[1] <= 12 and d[1] >= 1]

        # The last date belongs to the first panorama
        year, month = dates.pop(-1)
        pans[0].update({'year': year, "month": month})

        # The dates then apply in reverse order to the bottom panoramas
        dates.reverse()
        for i, (year, month) in enumerate(dates):
            pans[-1-i].update({'year': year, "month": month})

    if disp:
        for pan in pans:
            print(pan)

    if closest and len(dates) > 0:
        return [pans[i] for i in range(len(dates))]
    else:
        return pans



import os
import random
import time
import datetime
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
import warnings
import requests
import cv2
import pkg_resources
from pathlib import Path
import geopandas as gpd
from tqdm import tqdm
from shapely.geometry import Point
import warnings
from shapely.errors import ShapelyDeprecationWarning
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
import csv 

from zensvi.download.utils.imtool import ImageTool
from zensvi.download.utils.get_pids import panoids
from zensvi.download.utils.geoprocess import GeoProcessor
from zensvi.download.utils.helpers import standardize_column_names, create_buffer_gdf

class StreetViewDownloader:
    def __init__(self, gsv_api_key = None, log_path = None, nthreads = 0, distance = 1, grid = False, grid_size = 1):
        if gsv_api_key == None:
            warnings.warn("Please provide your Google Street View API key to augment metadata.")
        self._gsv_api_key = gsv_api_key
        self._log_path = log_path
        self._nthreads = nthreads
        self._distance = distance
        self._user_agent = self._get_ua()
        self._proxies = self._get_proxies()
        self._grid = grid
        self._grid_size = grid_size

    @property
    def gsv_api_key(self):
        return self._gsv_api_key    
    @gsv_api_key.setter
    def gsv_api_key(self,gsv_api_key):
        self._gsv_api_key = gsv_api_key

    @property
    def log_path(self):
        return self._log_path    
    @log_path.setter
    def log_path(self,log_path):
        self._log_path = log_path
        
    @property
    def nthreads(self):
        return self._nthreads    
    @nthreads.setter
    def nthreads(self,nthreads):
        self._nthreads = nthreads
    
    @property
    def distance(self):
        return self._distance    
    @distance.setter
    def distance(self,distance):
        self._distance = distance
    
    @property
    def grid(self):
        return self._grid    
    @grid.setter
    def grid(self,grid):
        self._grid = grid
        
    @property
    def grid_size(self):
        return self._grid_size    
    @grid_size.setter
    def grid_size(self,grid_size):
        self._grid_size = grid_size
    
    @property
    def user_agent(self):
        return self._user_agent  
    
    def _get_ua(self):
        user_agent_file = pkg_resources.resource_filename('zensvi.download.utils', 'UserAgent.csv')
        UA = []
        with open(user_agent_file, 'r') as f:
            for line in f:
                ua = {"user_agent": line.strip()}
                UA.append(ua)
        return UA
    
    @property
    def proxies(self):
        return self._proxies
    
    def _get_proxies(self):
        proxies_file = pkg_resources.resource_filename('zensvi.download_async.utils', 'proxies.csv')
        proxies = []
        with open(proxies_file, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                ip = row['ip']
                port = row['port']
                protocols = row['protocols']
                proxy_dict = {protocols: f"{ip}:{port}"}
                proxies.append(proxy_dict)
        return proxies
    
    def _read_pids(self, path_pid):
        pid_df = pd.read_csv(path_pid)
        # get unique pids as a list
        pids = pid_df.iloc[:,0].unique().tolist()
        return pids

    def _check_already(self, all_panoids):
        name_r, all_panoids_f = set(), []
        for name in os.listdir(self.dir_output):
            name_r.add(name.split(".")[0])

        for pid in all_panoids:
            if pid not in name_r:
                all_panoids_f.append(pid)
        return all_panoids_f

    def _log_write(self, pids):
        with open(self.log_path, 'a+') as fw:
            for pid in pids:
                fw.write(pid+'\n')
    
    def _augment_metadata(self, df):
        def get_year_month(pid, proxies):
            url = "https://maps.googleapis.com/maps/api/streetview/metadata?pano={}&key={}".format(pid, self.gsv_api_key)
            while True:
                proxy = random.choice(proxies)
                try:
                    response = requests.get(url, proxies=proxy, timeout=5)
                    break
                except Exception as e:
                    print(f"Proxy {proxy} is not working. Exception: {e}")
                    continue
            response = response.json()
            if response['status'] == 'OK':
                # get year and month from date
                try:
                    date = response['date']
                    year = date.split("-")[0]
                    month = date.split("-")[1]
                except Exception:
                    year = None
                    month = None
                return {"year": year, "month": month}
            return {"year": None, "month": None}    

        def worker(row, proxies):
            panoid = row.panoid
            year_month = get_year_month(panoid, proxies)
            return row.Index, year_month
        
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size

        for i in tqdm(range(num_batches), desc=f"Augmenting metadata by batch size {min(batch_size, len(df))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row, self.proxies): row.Index for row in df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Augmenting metadata for batch #{i+1}"):
                    row_index, year_month = future.result()
                    df.at[row_index, 'year'] = year_month['year']
                    df.at[row_index, 'month'] = year_month['month']
        return df
                    
    def _get_pids_from_csv(self, df, id_columns, closest=False, disp=False):
        def get_street_view_info(longitude, latitude, proxies):
            results = panoids(latitude, longitude, proxies, closest=closest, disp=disp)
            return results

        def worker(row):
            input_longitude = row.longitude
            input_latitude = row.latitude
            return {column: getattr(row, column) for column in id_columns}, (input_longitude, input_latitude), get_street_view_info(input_longitude, input_latitude, self.proxies)

        results = []
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size
        failed_rows = []

        for i in tqdm(range(num_batches), desc=f"Getting pids by batch size {min(batch_size, len(df))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row): row for row in df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting pids for batch #{i+1}"):
                    try:
                        id_dict, (input_longitude, input_latitude), row_results = future.result()
                        for result in row_results:
                            result['input_longitude'] = input_longitude
                            result['input_latitude'] = input_latitude
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Error: {e}")
                        failed_rows.append(batch_futures[future])  # Store the failed row

        # Retry failed rows
        if failed_rows:
            print("Retrying failed rows...")
            with ThreadPoolExecutor() as executor:
                retry_futures = {executor.submit(worker, row): row for row in failed_rows}
                for future in tqdm(as_completed(retry_futures), total=len(retry_futures), desc="Retrying failed rows"):
                    try:
                        id_dict, (input_longitude, input_latitude), row_results = future.result()
                        for result in row_results:
                            result['input_longitude'] = input_longitude
                            result['input_latitude'] = input_latitude
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Failed again: {e}")

        results_df = pd.DataFrame(results)
        return results_df

    def _get_pids_from_gdf(self, gdf, id_columns, closest=False, disp=False):  
        # read shapefile
        gp = GeoProcessor(gdf, distance=self.distance, grid=self.grid, grid_size=self.grid_size, id_columns = id_columns)
        df = gp.get_lat_lon()

        # Use _get_pids_from_csv to get pids from df
        results_df = self._get_pids_from_csv(df, id_columns, closest=closest, disp=disp)

        # Check if lat and lon are within input polygons
        polygons = gpd.GeoSeries([geom for geom in gdf['geometry'] if geom.type in ['Polygon', 'MultiPolygon']])

        # Convert lat, lon to Points and create a GeoSeries
        points = gpd.GeoSeries([Point(lon, lat) for lon, lat in zip(results_df['lon'], results_df['lat'])])

        # Create a GeoDataFrame with the points and an index column
        points_gdf = gpd.GeoDataFrame(geometry=points, crs=gdf.crs)
        points_gdf['index'] = range(len(points_gdf))

        # Create a spatial index on the polygons GeoSeries
        polygons_sindex = polygons.sindex

        # Function to check whether a point is within any polygon
        def is_within_polygon(point):
            possible_matches_index = list(polygons_sindex.intersection(point.bounds))
            possible_matches = polygons.iloc[possible_matches_index]
            precise_matches = possible_matches.contains(point)
            return precise_matches.any()

        # Add progress bar for within_polygon calculation
        with tqdm(total=len(points), desc="Checking points within polygons") as pbar:
            within_polygon = []
            for point in points_gdf['geometry']:
                within_polygon.append(is_within_polygon(point))
                pbar.update()

        results_df['within_polygon'] = within_polygon

        # Return only those points within polygons
        results_within_polygons_df = results_df[results_df['within_polygon']]
        # Drop the 'within_polygon' column
        results_within_polygons_df = results_within_polygons_df.drop(columns='within_polygon')
        return results_within_polygons_df


    def get_pids(self, path_pid, lat = None, lon = None, input_csv_file = "", input_shp_file = "", id_columns=None, buffer = 0, closest=False, disp=False, augment_metadata=False):
        if id_columns is not None:
            if isinstance(id_columns, str):
                id_columns = [id_columns.lower()]
            elif isinstance(id_columns, list):
                id_columns = [column.lower() for column in id_columns]
        else:
            id_columns = []
        if lat != None and lon != None:
            pid = panoids(lat, lon, closest=closest, disp=disp)
        elif input_csv_file != "":
            df = pd.read_csv(input_csv_file)
            df = standardize_column_names(df)
            if buffer > 0:
                gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.longitude, df.latitude), crs='EPSG:4326')
                gdf = create_buffer_gdf(gdf, buffer)
                pid = self._get_pids_from_gdf(gdf, id_columns, closest=False, disp=False)
            else:
                pid = self._get_pids_from_csv(df, id_columns, closest=False, disp=False)
        elif input_shp_file != "":
            gdf = gpd.read_file(input_shp_file)
            if buffer > 0:
                gdf = create_buffer_gdf(gdf, buffer)
            pid = self._get_pids_from_gdf(gdf, id_columns, closest=False, disp=False)
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")
        # save the pids
        pid_df = pd.DataFrame(pid)
        pid_df = pid_df.drop_duplicates(subset='panoid')
        if augment_metadata & (self.gsv_api_key != None):
            pid_df = self._augment_metadata(pid_df)
        elif augment_metadata & (self.gsv_api_key == None):
            raise ValueError("Please set the gsv api key by calling the gsv_api_key method.")
        pid_df.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid))
    
    def download_gsv(self, dir_output, path_pid = None, zoom=2, h_tiles=4, v_tiles=2, cropped=False, full=True, 
                lat=None, lon=None, input_csv_file="", input_shp_file = "", id_columns=None, buffer = 0, closest=False, disp=False, augment_metadata=False):
        # set dir_output as attribute and create the directory
        self.dir_output = dir_output
        Path(dir_output).mkdir(parents=True, exist_ok=True)
        
        # call get_pids function first if path_pid is None
        if path_pid is None:
            print("Getting pids...")
            path_pid = os.path.join(self.dir_output, "pids.csv")
            self.get_pids(path_pid, lat=lat, lon=lon,
                        input_csv_file=input_csv_file, input_shp_file = input_shp_file, id_columns=id_columns, buffer = buffer, closest=closest, disp=disp, augment_metadata=augment_metadata)


        # Horizontal Google Street View tiles
        # zoom 3: (8, 4); zoom 5: (26, 13) zoom 2: (4, 2) zoom 1: (2, 1);4:(8,16)
        # zoom = 2
        # h_tiles = 4  # 26
        # v_tiles = 2  # 13
        # cropped = False
        # full = True
        # create a folder within self.dir_output
        panorama_output = os.path.join(self.dir_output, "panorama")
        os.makedirs(panorama_output, exist_ok=True)
        
        panoids = self._read_pids(path_pid)

        if len(panoids) == 0:
            print("There is no panorama ID to download.")
            return
        else:
            panoids_rest = self._check_already(panoids)

        if len(panoids_rest) > 0:
            UAs = random.choices(self.user_agent, k = len(panoids_rest))
            ImageTool.dwl_multiple(panoids_rest, zoom, v_tiles, h_tiles, panorama_output, UAs, self.proxies, cropped, full, log_path=self.log_path)
        else:
            print("All images have been downloaded.")